//******************************************************************************
//
// File:    GenKey.java
// Package: ---
// Unit:    Class GenKey
//
// This Java source file is copyright (C) 2016 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.util.Random;
import edu.rit.util.RandomSubset;

/**
 * Class GenKey is a main program that generates a random key for the LC4
 * cipher.
 * <P>
 * Usage: <TT>java GenKey [<I>seed</I>]</TT>
 *
 * @author  Alan Kaminsky
 * @version 26-May-2016
 */
public class GenKey
	{

// Prevent construction.

	private GenKey()
		{
		}

// Main program.

	/**
	 * Main program.
	 */
	public static void main
		(String[] args)
		{
		if (args.length > 1) usage();
		long seed =
			(args.length == 1) ? Long.parseLong (args[0]) :
				System.currentTimeMillis();
		Random prng = new Random (seed);
		RandomSubset rs = new RandomSubset (prng, 36, true);
		int[] key = new int [36];
		Util.genKey (key, rs);
		for (int i = 0; i < 6; ++ i)
			{
			for (int j = 0; j < 6; ++ j)
				System.out.printf (" %c", Util.intToLetter (key[6*i+j]));
			System.out.println();
			}
		for (int i = 0; i < 36; ++ i)
			System.out.printf ("%c", Util.intToLetter (key[i]));
		System.out.println();
		}

// Hidden operations.

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java GenKey [<seed>]");
		System.exit (1);
		}

	}
