//******************************************************************************
//
// File:    LC4.java
// Package: ---
// Unit:    Class LC4
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.util.BitSet64;

/**
 * Class LC4 implements the ElsieFour low-tech cipher.
 * <P>
 * <I>Note:</I> Class LC4 is not multiple thread safe.
 *
 * @author  Alan Kaminsky
 * @version 28-Mar-2017
 * 
 * modified by Anna Lena Rotthaler
 * 01-Oct-2017
 * Weakened version without nonce added for encryption and decryption
 */
public class LC4
	{

// Exported data members.

	/**
	 * True to print a trace during encryption or decryption.
	 */
	public static boolean trace = false;

// Hidden data members.

	private static final int N = 6;
	private static final int NSQR = N*N;

	private int[][] state = new int [N] [N];
	private int kval, krow, kcol;
	private int pval, prow, pcol;
	private int cval, crow, ccol;
	private BitSet64 bitset = new BitSet64();

// Exported constructors.

	/**
	 * Construct a new integer LC4 cipher.
	 */
	public LC4()
		{
		}

// Exported operations.

	/**
	 * Encrypt the given message. The key must be a permutation of the integers
	 * 0 to 35. The nonce and plaintext message must consist of integers 0 to
	 * 35.
	 *
	 * @param  key    Key (input).
	 * @param  nonce  Nonce (input).
	 * @param  ptmsg  Plaintext message (input).
	 * @param  ctmsg  Ciphertext message (output).
	 *
	 * @exception  IllegalArgumentException
	 *     (unchecked exception) Thrown if <TT>key</TT>, <TT>nonce</TT>, or
	 *     <TT>ptmsg</TT> does not meet the above preconditions.
	 */
	public void encrypt
		(int[] key,
		 int[] nonce,
		 int[] ptmsg,
		 int[] ctmsg)
		{
		start (verifyKey (key));
		doEncrypt (verify (nonce), null);
		doEncrypt (verify (ptmsg), ctmsg);
		}
	
	public void encrypt (int[] key, int[] ptmsg, int[] ctmsg)
	{
		start(verifyKey (key));
		doEncrypt(verify(ptmsg), ctmsg);
	}

	/**
	 * Decrypt the given message. The key must be a permutation of the integers
	 * 0 to 35. The nonce and ciphertext message must consist of integers 0 to
	 * 35.
	 *
	 * @param  key    Key.
	 * @param  nonce  Nonce.
	 * @param  ctmsg  Ciphertext message (input).
	 * @param  ptmsg  Plaintext message (output).
	 *
	 * @exception  IllegalArgumentException
	 *     (unchecked exception) Thrown if <TT>key</TT>, <TT>nonce</TT>, or
	 *     <TT>ctmsg</TT> does not meet the above preconditions.
	 */
	public void decrypt
		(int[] key,
		 int[] nonce,
		 int[] ctmsg,
		 int[] ptmsg)
		{
		start (verifyKey (key));
		doEncrypt (verify (nonce), null);
		doDecrypt (verify (ctmsg), ptmsg);
		}

	public void decrypt
		(int[] key,
		 int[] ctmsg,
		 int[] ptmsg)
		{
		start (verifyKey (key));
		doDecrypt (verify (ctmsg), ptmsg);
		}
		
	/**
	 * Encrypt the given message. The key must be a permutation of the
	 * characters #, _, 2 to 9, and A to Z (case insensitive). The nonce and
	 * plaintext message must consist of the characters #, _, 2 to 9, and A to Z
	 * (case insensitive).
	 *
	 * @param  key    Key.
	 * @param  nonce  Nonce.
	 * @param  ptmsg  Plaintext message.
	 *
	 * @return  Ciphertext message.
	 */
	public String encrypt
		(String key,
		 String nonce,
		 String ptmsg)
		{
		int[] ctmsg = new int [ptmsg.length()];
		encrypt
			(Util.lettersToInts (key),
			 Util.lettersToInts (nonce),
			 Util.lettersToInts (ptmsg),
			 ctmsg);
		return Util.intsToLetters (ctmsg);
		}

	public String encrypt
		(String key,
		 String ptmsg)
		{
		int[] ctmsg = new int [ptmsg.length()];
		encrypt
			(Util.lettersToInts (key),
			 Util.lettersToInts (ptmsg),
			 ctmsg);
		return Util.intsToLetters (ctmsg);
		}
		
	/**
	 * Decrypt the given message. The key must be a permutation of the
	 * characters #, _, 2 to 9, and A to Z (case insensitive). The nonce and
	 * ciphertext message must consist of the characters #, _, 2 to 9, and A to
	 * Z (case insensitive).
	 *
	 * @param  key    Key.
	 * @param  nonce  Nonce.
	 * @param  ctmsg  Ciphertext message.
	 *
	 * @return  Plaintext message.
	 */
	public String decrypt
		(String key,
		 String nonce,
		 String ctmsg)
		{
		int[] ptmsg = new int [ctmsg.length()];
		decrypt
			(Util.lettersToInts (key),
			 Util.lettersToInts (nonce),
			 Util.lettersToInts (ctmsg),
			 ptmsg);
		return Util.intsToLetters (ptmsg);
		}

	public String decrypt
		(String key,
		 String ctmsg)
		{
		int[] ptmsg = new int [ctmsg.length()];
		decrypt
			(Util.lettersToInts (key),
			 Util.lettersToInts (ctmsg),
			 ptmsg);
		return Util.intsToLetters (ptmsg);
		}
		
// Hidden operations.

	/**
	 * Verify preconditions for the given message.
	 *
	 * @param  msg  Message.
	 *
	 * @return  Verified message.
	 *
	 * @exception  IllegalArgumentException
	 *     (unchecked exception) Thrown if any element of <TT>msg</TT> is not an
	 *     integer from 0 to 35.
	 */
	private static int[] verify
		(int[] msg)
		{
		for (int i = 0; i < msg.length; ++ i)
			{
			if (0 > msg[i] || msg[i] > 35)
				throw new IllegalArgumentException (String.format
					("LC4: Element %d illegal", msg[i]));
			}
		return msg;
		}

	/**
	 * Verify preconditions for the given key.
	 *
	 * @param  key  Key.
	 *
	 * @return  Verified message.
	 *
	 * @exception  IllegalArgumentException
	 *     (unchecked exception) Thrown if <TT>key</TT> is not a permutation of
	 *     the integers from 0 to 24.
	 */
	private int[] verifyKey
		(int[] key)
		{
		bitset.clear();
		for (int i = 0; i < key.length; ++ i)
			{
			if (0 > key[i] || key[i] > 35)
				throw new IllegalArgumentException (String.format
					("LC4: Element %d illegal", key[i]));
			if (bitset.contains (key[i]))
				throw new IllegalStateException
					("LC4: Key is not a valid permutation");
			bitset.add (key[i]);
			}
		if (bitset.size() != NSQR)
			throw new IllegalStateException
				("LC4: Key is not a valid permutation");
		return key;
		}

	/**
	 * Begin an encryption or decryption. The cipher state is initialized.
	 *
	 * @param  key  Key.
	 */
	private void start
		(int[] key)
		{
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				state[i][j] = key[N*i+j];
		krow = kcol = 0;
		}

	/**
	 * Perform the encryption algorithm.
	 *
	 * @param  ptmsg  Plaintext message (input).
	 * @param  ctmsg  Ciphertext message (output).
	 */
	private void doEncrypt
		(int[] ptmsg,
		 int[] ctmsg)
		{
		if (trace)
			{
			System.out.printf ("doEncrypt (ptmsg[%d], ", ptmsg.length);
			if (ctmsg == null)
				System.out.printf ("null");
			else
				System.out.printf ("ctmsg[%d]", ctmsg.length);
			System.out.printf (")%n");
			System.out.printf ("    ");
			printState();
			System.out.println();
			}

		for (int i = 0; i < ptmsg.length; ++ i)
			{
			pval = ptmsg[i];
			prow = getRowIndex (pval);
			pcol = getColIndex (pval);

			kval = state[krow][kcol];

			crow = increaseRowIndex (prow, kval);
			ccol = increaseColIndex (pcol, kval);
			cval = state[crow][ccol];
			if (ctmsg != null) ctmsg[i] = cval;

			shiftRow (prow);
			shiftCol (ccol);

			krow = increaseRowIndex (krow, cval);
			kcol = increaseColIndex (kcol, cval);

			if (trace)
				{
				System.out.printf ("%-3d ", i);
				printState();
				System.out.printf ("%-3c%-3c%n",
					Util.intToLetter (pval), Util.intToLetter (cval));
				}
			}
		}

	/**
	 * Perform the decryption algorithm.
	 *
	 * @param  ctmsg  Ciphertext message (input).
	 * @param  ptmsg  Plaintext message (output).
	 */
	private void doDecrypt
		(int[] ctmsg,
		 int[] ptmsg)
		{
		if (trace)
			{
			System.out.printf ("doDecrypt (ctmsg[%d], ", ctmsg.length);
			if (ptmsg == null)
				System.out.printf ("null");
			else
				System.out.printf ("ptmsg[%d]", ptmsg.length);
			System.out.printf (")%n");
			System.out.printf ("    ");
			printState();
			System.out.println();
			}

		for (int i = 0; i < ctmsg.length; ++ i)
			{
			cval = ctmsg[i];
			crow = getRowIndex (cval);
			ccol = getColIndex (cval);

			kval = state[krow][kcol];

			prow = decreaseRowIndex (crow, kval);
			pcol = decreaseColIndex (ccol, kval);
			pval = state[prow][pcol];
			if (ptmsg != null) ptmsg[i] = pval;

			shiftRow (prow);
			shiftCol (ccol);

			krow = increaseRowIndex (krow, cval);
			kcol = increaseColIndex (kcol, cval);

			if (trace)
				{
				System.out.printf ("%-3d ", i);
				printState();
				System.out.printf ("%-3c%-3c%n",
					Util.intToLetter (cval), Util.intToLetter (pval));
				}
			}
		}

	/**
	 * Returns the state row index in which value v appears.
	 */
	private int getRowIndex
		(int v)
		{
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				if (state[i][j] == v)
					return i;
		throw new IllegalStateException (String.format
			("LC4.getRowIndex(%d): Shouldn't happen", v));
		}

	/**
	 * Returns the state column index in which value v appears.
	 */
	private int getColIndex
		(int v)
		{
		for (int i = 0; i < N; ++ i)
			for (int j = 0; j < N; ++ j)
				if (state[i][j] == v)
					return j;
		throw new IllegalStateException (String.format
			("LC4.getColIndex(%d): Shouldn't happen", v));
		}

	/**
	 * Returns row index r increased as determined by value v.
	 */
	private static int increaseRowIndex
		(int r,
		 int v)
		{
		return (r + v/N)%N;
		}

	/**
	 * Returns row index r decreased as determined by value v.
	 */
	private static int decreaseRowIndex
		(int r,
		 int v)
		{
		return (r - v/N + N)%N;
		}

	/**
	 * Returns column index c increased as determined by value v.
	 */
	private static int increaseColIndex
		(int c,
		 int v)
		{
		return (c + v%N)%N;
		}

	/**
	 * Returns column index c decreased as determined by value v.
	 */
	private static int decreaseColIndex
		(int c,
		 int v)
		{
		return (c - v%N + N)%N;
		}

	/**
	 * Shift state row at index r one place right. Also update indexes.
	 */
	private void shiftRow
		(int r)
		{
		int t = state[r][N-1];
		for (int c = N - 1; c > 0; -- c)
			state[r][c] = state[r][c-1];
		state[r][0] = t;
		if (krow == r) kcol = incrIndex (kcol);
		if (prow == r) pcol = incrIndex (pcol);
		if (crow == r) ccol = incrIndex (ccol);
		}

	/**
	 * Shift state column at index c one place down.
	 */
	private void shiftCol
		(int c)
		{
		int t = state[N-1][c];
		for (int r = N - 1; r > 0; -- r)
			state[r][c] = state[r-1][c];
		state[0][c] = t;
		if (kcol == c) krow = incrIndex (krow);
		if (pcol == c) prow = incrIndex (prow);
		if (ccol == c) crow = incrIndex (crow);
		}

	/**
	 * Returns row or column index x incremented.
	 */
	private static int incrIndex
		(int x)
		{
		return (x + 1)%N;
		}

	/**
	 * Print the state.
	 */
	private void printState()
		{
		for (int i = 0; i < N; ++ i)
			{
			for (int j = 0; j < N; ++ j)
				System.out.printf ("%c", Util.intToLetter (state[i][j]));
			System.out.printf (" ");
			}
		System.out.printf (" %-3d%-3d", krow, kcol);
		}

	}
